# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals

from django_sorcery.db import signals  # noqa

from ..base import TestCase


class TestSession(TestCase):

    def test_session_signals(self):
        pass
